#define _GNU_SOURCE // O_DIRECT
#define _XOPEN_SOURCE 600 // pread > 500, posix_memalign > 600
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "list.h"
#include "trace.h"

#define stats trace_on
#define verbose trace_off
#define trace trace_off
#define names trace_off
#define tags trace_off
#define tracediff trace_off
#define tracedump trace_off
#define cruft trace_off
#define lcount trace_on
#define ltrace trace_on
#define ltdots 1

typedef unsigned char uchar;
typedef int fd_t;

static void hexdump(void *data, unsigned len)
{
	while (len) {
		int line = 16, n = len < line? len: line, i;
		printf("%p: ", data);
		for (i = 0; i < n; i++)
			printf("%02hhx ", *(uchar *)data++);
		printf("%*s", 3*(line - n) + 2, "");
		for (data -= n, i = 0; i < n; i++) {
			unsigned char c = *(uchar *)data++;
			printf("%c", c < ' ' ? '.' : c);
		}
		printf("\n");
		data += n;
		len -= n;
	}
}

static void memcpz(char *to, char *from, unsigned len)
{
	memcpy(to, from, len);
	to[len] = 0;
}

static void memcpc(void *to, void *from, unsigned len)
{
	*(char *)to = len;
	memcpy(to + 1, from, len);
}

static unsigned stringhash(const unsigned char *str, unsigned len)
{
	unsigned hash0 = 0x12a3fe2d, hash1 = 0x37a6e8f9;
	while (len--) {
		unsigned hash = hash1 + (hash0 ^ (*str++ * 0x7d9e8da1));
		if (hash & 0x80000000) hash -= 0x7fffffff;
			hash1 = hash0;
			hash0 = hash;
		}
		return hash0;
}

static int getu(uchar *ptr, unsigned len, unsigned *val)
{
	unsigned sum = 0;
	if (!len)
		return -2;
	while (len--) {
		char c = *ptr++;
		if (!isdigit(c))
			return -1;
		sum = sum*10 + (c - '0');
	}	
	*val = sum;
	return 0;
}

struct cstring { unsigned char length; unsigned char text[]; };

static struct cstring *cstring(char *str, unsigned len)
{
	struct cstring *new = malloc(len + 1);
	memcpc(new, str, len);
	return new;
}

static struct cstring *cstrz(char *strz)
{
	return cstring(strz, strlen(strz));
}

static void print_cstring(struct cstring *cstring)
{
	printf("%.*s ", cstring->length, cstring->text);
}

static int csless(struct cstring *a, struct cstring *b)
{
	int len1 = a->length, len2 = b->length, less;
	// printf("is %.*s less than %.*s? ", len1, a->text, len2, b->text);
	if (!(less = memcmp(a->text, b->text, len1 < len2? len1: len2)))
		return len1 < len2;
	return less < 0;
}

static struct list_head *new_hashvec(unsigned buckets)
{
	struct list_head *hash = malloc(buckets * sizeof(struct list_head));
	int i;
	for (i = 0; i < buckets; i++)
		INIT_LIST_HEAD(hash + i);
	return hash;
}

static void relocate(int rel, ...)
{
	va_list args;
	va_start(args, rel);
	void **p;
	while((p = va_arg(args, void **)))
		*p += rel;
	va_end(args);
}

#define COMBSORT(SIZE, I, J, LESS, SWAP) { \
	unsigned gap = SIZE, more, I; \
	do { \
		if (gap > 1) gap = gap*10/13; \
		if (gap - 9 < 2) gap = 11; \
		for (more = gap > 1, I = SIZE - 1; I >= gap; I--) { \
			int J = I - gap; \
			if (LESS) { SWAP; more = 1; } } \
	} while (more); }

/* Generic memory block list */

struct memvec { 
	struct cstring *id;
	unsigned items, maxitems, blocks, maxblocks, itemsize, perblock;
	char *block[];
};

static int expand(unsigned size, void *memvar)
{
	void *new = realloc(*(void **)memvar, size);
	return new? ((*(void **)memvar = new), 0): ((errno = ENOMEM), -1);
}

static int expand_aligned(void **memvar, unsigned *size, unsigned more, unsigned order)
{
	void *old = *memvar;
	if (posix_memalign(memvar, 1 << order, *size + more))
		error("memalign error %s (%u)", strerror(errno), errno);
	memcpy(*memvar, old, *size);
	*size += more;
	free(old);
	return 0;
}

/* itemvec: single-level expanding item storage */

static void *new_vecitem(void **vecref, unsigned head, unsigned itemsize, unsigned more)
{
	struct itemvec { unsigned maxitems, items; char item[]; } *vec = *vecref + head;
	trace_off(warn("check vec %p, items = %u maxitems = %u", vec, vec->items, vec->maxitems);)
	if (vec->items == vec->maxitems) {
		unsigned expanded = vec->items + more;
		trace_off(warn("expand vec to %u", expanded);)
		expand(head + sizeof(struct itemvec) + itemsize * expanded, vecref);
		vec = *vecref + head;
		vec->maxitems = expanded;
	}
	return vec->item + itemsize * vec->items++;
}

/* memvec: two-level expanding item storage */

static struct memvec *new_memvec(unsigned perblock, unsigned itemsize, char *id)
{
	struct memvec *new = malloc(sizeof(struct memvec));
	if (new)
		*new = (struct memvec){ .perblock = perblock, .itemsize = itemsize, .id = cstrz(id) };
	return new;
}

static void *new_memitem(struct memvec **memvar)
{
	struct memvec *mem = *memvar;
	if (mem->items != mem->maxitems) {
		unsigned last = mem->blocks - 1;
		unsigned offset = mem->items++ - last*mem->perblock;
		return mem->block[last] + offset*mem->itemsize;
	}
	if (mem->blocks == mem->maxblocks) {
		mem->maxblocks += mem->perblock;
		if (expand(sizeof(struct memvec) + mem->maxblocks*sizeof(char *), &mem))
			return NULL;
		*memvar = mem;
	}
	void *new;
	unsigned more = mem->perblock;
	if (!(new = malloc(more * mem->itemsize)))
		return NULL;
	mem->block[mem->blocks++] = new;
	mem->maxitems += more;
//	trace(warn("new %.*s block, %u-%u", mem->name->length, mem->name->text, mem->items, mem->maxitems);)
	mem->items++;
	return new;
}

static void walk_memvec(struct memvec *mem, void (*fn)(void *), int seqnums)
{
	int i, j, m = mem->perblock, n = mem->items;
	for (i = 0; i < mem->blocks; i++, n -=m) {
		void *item = mem->block[i];
		for (j = 0; j < (m < n? m: n); j++, item += mem->itemsize) {
			if (seqnums)
				printf("%u: ", i*m + j);
			(fn)(item);
		}
	}
}

static void show_memvec(char *label, struct memvec *mem)
{
	int i, n = mem? mem->blocks: 0;
	printf("%u %s blocks: \n", n, label);
	for (i = 0; i < n; i++) {
		printf("%p ", mem->block[i]);
	}
	printf("\n");
}

static void *index_item(unsigned id, struct memvec *mem)
{
	assert(id < mem->items);
	unsigned perblock = mem->perblock, block = id / perblock, offset = id - block * perblock;
	return mem->block[block] + offset*mem->itemsize;
}

/* Input line count */

static unsigned inlines = 0;

static void show_line(void) {
	if (!inlines || (inlines % 100000))
		return;
	if (ltdots) { 
		fputc('.', stderr);
		fsync(1);
		return;
	}
	printf("[%u]\n", inlines);
}

unsigned line(void)
{
	return inlines + 1;
}

/* Text file parsing */

struct text {
	fd_t file, direct;
	off_t filesize, filetail;
	uchar *base, *item, *at, *lim, *top;
	unsigned size, chunk, maxline, align;
};

static off_t filepos(struct text *text)
{
	return text->filesize - text->filetail;
}

static void show_pos(struct text *text)
{
	trace(warn("at %x+%x, lim=%x top=%x size=%x more=%Lx",
		text->item - text->base,
		text->at - text->item,
		text->lim - text->base,
		text->top - text->base,
		text->chunk + text->maxline,
		(long long)text->filetail);)
}

static void get_more(struct text *text)
{
	trace(printf("\n");)
	assert(text->base <= text->item);
	assert(text->item <= text->at);
	unsigned alignmask = (1 << text->align) - 1;
	unsigned keep = text->top - text->item;
	unsigned pad = (text->item - text->base) & alignmask;
	unsigned rel = text->base + pad - text->item;
	trace(warn("keep = %x, pad = %x, new top = %x", keep, pad, keep + pad);)

	if (rel) {
		trace(warn("move %x bytes from %x to %x", keep, text->item - text->base, pad);)
		memmove(text->base + pad, text->item, keep);
		relocate(rel, &text->item, &text->at, &text->top, &text->lim, 0);
		trace(show_pos(text);)
	}

	unsigned room = text->size - ((keep + pad) & ~alignmask);
	unsigned grab = text->filetail <= room? text->filetail: room;
	trace_off(show_pos(text);)
	trace_off(warn("grab %u more bytes", grab);)
	if (!rel && !grab && text->filetail > room) {
		trace(warn("expand buffer to %u!!", text->size);)
		uchar *old = text->base;
		if (1)
			expand_aligned((void **)&text->base, &text->size, text->chunk, text->align);
		else
			expand(text->size, &text->base);
		relocate(text->base - old, &text->item, &text->at, &text->top, &text->lim, 0);
		trace_on(show_pos(text);)
		grab = text->chunk;
	}

	trace(warn("load %x bytes to %x", grab, (keep + pad));)
	text->top = text->base + (keep + pad);
	if (grab) {
		fd_t fd = filepos(text) < 800*1000*1000? text->file: text->direct;
		unsigned got = pread(fd, text->base + (keep + pad), grab + (-grab & alignmask), filepos(text));
		if (got == -1)
			error("read error %s (%u)", strerror(errno), errno);
		trace(warn("loaded %u bytes", got);)
		text->filetail -= got;
		text->top += got;
	}
	text->lim = text->top - text->at < text->maxline? text->top: text->top - text->maxline;
	trace(show_pos(text);)
	assert(text->at <= text->lim);
	assert(text->item <= text->lim);
	assert(text->lim <= text->top);
	assert(text->at <= text->top);
}

static inline void need_more(struct text *text)
{
	if (text->at >= text->lim)
		get_more(text);
}

/* deltas, udiff parsing */

#define text_op 0
#define copy_op 1
#define skip_op 2
#define high_op 3

#define text(n) (n | (text_op << 6))
#define copy(n) (n | (copy_op << 6))
#define skip(n) (n | (skip_op << 6))
#define move(n, s) copy(0), copy(s), copy(n)

static int alter(uchar *ops, uchar *in, uchar *out)
{
	unsigned c, state = 0, count = 0;
	uchar *org = in, *source = 0;

more:	count = 0;
	while ((c = *ops++)) {
		count = (count << 6) | (c & 0x3f);
		switch (c >> 6) {
		case text_op:
			memcpy(out, ops, count);
			out +=count;
			ops += count;
			goto more;
		case copy_op:
			if (!count)
				goto move;
			memcpy(out, in, count);
			out +=count;
		case skip_op:
			in +=count;
			goto more;
		}
	}
	goto done;

move:	count = 0;
	while ((c = *ops++)) {
		count = (count << 6) | (c & 0x3f);
		switch (c >> 6) {
		case copy_op:
			if (state) {
				memcpy(out, source, count);
				out += count;
				state = 0;
				goto more;
			}
			source = org + count;
			state = 1;
			goto move;
		case text_op:
		case skip_op:
			goto bork;
		}
	}
done:	return 0;
bork:	return -1;
}

// !!! should not allow leading 0 for high
// !!! should trap all invalid opcodes
// !!! should take ops string limit and check against it
static struct transinfo {int in; int out;} transcheck(uchar *ops)
{
	unsigned c, state = 0, count = 0, ilen = 0, olen = 0;

more:	count = 0;
	while ((c = *ops++)) {
		count = (count << 6) | (c & 0x3f);
		switch (c >> 6) {
		case text_op:
			olen +=count;
			ops += count;
			goto more;
		case copy_op:
			if (!count)
				goto move;
			olen +=count;
		case skip_op:
			ilen +=count;
			goto more;
		}
	}
	goto done;

move:	count = 0;
	while ((c = *ops++)) {
		count = (count << 6) | (c & 0x3f);
		switch (c >> 6) {
		case copy_op:
			if (state) {
				olen += count;
				state = 0;
				goto more;
			}
			state = 1;
			goto move;
		case text_op:
		case skip_op:
			goto bork;
		}
	}
done:	return (struct transinfo) {ilen, olen};
bork:	return (struct transinfo) {-1, 0};
}

// statics for generate...
// should not be static, use a struct
// haven't tested op len > 64

static struct textline { uchar *source; unsigned length; } *textlines; // check limits!!!

static uchar *emit_text, *this_text, *end_text;
static int emit_line, this_line, hold_lines, hold_length, emit_op, maxlines;
static uchar *outbuf, *output, *outlim;
static unsigned delta = 0, diffchunk = 0;

static void get_more_diff(struct text *text)
{
	uchar *old = text->base;
	get_more(text);
	int rel = text->base - old, i;
	if (!rel)
		return;
	for (i = 0; i < hold_lines; i++)
		textlines[i].source += rel;
}

static inline void need_more_diff(struct text *text)
{
	if (text->at >= text->lim) {
		get_more_diff(text);
	}
}

static void expand_output(unsigned need)
{
	unsigned have = outlim - output, more = 1 << 12;
	if (need > have) {
		uchar *old = outbuf;
		unsigned bytes = need > have + more? need: more;
		unsigned size = outlim - outbuf + bytes;
		trace_off(warn("output size = %i, have = %u, need = %u", outlim - outbuf, have, need);)
		expand(size, &outbuf);
		relocate(outbuf - old, &output, 0);
		outlim = outbuf + size;
		trace_off(warn("expand output by %u to %u", bytes, outlim - outbuf);)
	}
}

tracediff(static int listing = 1;)
tracediff(static void printc(char c) { if (listing) putchar(c == '\n'? '\\': c); })
tracediff(static void prints(char *s, unsigned n) { if (n) while (n--) printc(*s++); })
tracediff(static void print1(char c) { printc(c); if (listing) printf("? "); })
tracediff(static void print2(char c) { printc(c); if (listing) printf(", "); })

static void lcount_text(struct text *text)
{
	if (*text->at == '\n') {
		inlines++;
		ltrace(show_line();)
	}
}

void unnext(struct text *text)
{
	--text->at;
	lcount(if (*text->at == '\n') --inlines;)
}

static inline int _next_is(struct text *text, uchar c)
{
	need_more_diff(text);
	if (text->at < text->top) {
		tracediff(if (listing) print1(*text->at);)
		lcount(lcount_text(text);)
		return *text->at++ == c;
	}
	return 0;
}

static inline int _skip_to(struct text *text, uchar c)
{
	while (text->at < text->top) {
		need_more_diff(text);
		tracediff(if (listing) print2(*text->at);)
		lcount(lcount_text(text);)
		if (*text->at++ == c)
			return 1;
	}
	return 0;
}

static inline char _next_ch(struct text *text)
{
	need_more_diff(text);
	if (text->at < text->top) {
		lcount(lcount_text(text);)
		return *text->at++;
	}
	return -1;
}

#define next_is(c) _next_is(text, (c))
#define skip_to(c) _skip_to(text, (c))
#define next_ch() _next_ch(text)

static void generate(uchar op)
{
	unsigned len = 0;
	struct text orig = { .file = -1, .base = this_text, .at = this_text, .lim = this_text, .top = this_text };

	switch (emit_op) {
	case copy_op:
	case skip_op:
		for (; emit_line < this_line; emit_line++)
			_skip_to(&orig, '\n');
		len = this_text - emit_text;
		break;
	case text_op:
		len = hold_length;
		break;
	}

	if (len) {
		if (len > 64) {
			int shift, any = 0;
			for (shift = 5*6; shift < 6; shift -= 6) {
				unsigned len1 = (len >> shift) & 0x3f;
				if ((any |= len1)) {
					expand_output(1);
					*output++ = (high_op << 6) | len1;
				}
			}
		}

		expand_output(1);
		*output++ = emit_op << 6 | (len & 0x3f);
		if (emit_op == text_op) {
			int i;
			for (i = 0; i < hold_lines; i++) {
				unsigned length = textlines[i].length;
				expand_output(length);
				memcpy(output, textlines[i].source, length);
				output += length;
			}
			hold_length = 0;
			hold_lines = 0;
		}
	}
	tracediff(if (listing) if (emit_op == skip_op) { printf("skip ["); prints(emit_text, this_text - emit_text); printf("]"); })
	tracediff(if (listing) if (emit_op == copy_op) { printf("copy ["); prints(emit_text, this_text - emit_text); printf("]"); })
	emit_op = op;
	emit_line = this_line; // should already be
	emit_text = this_text;
}

static int diff2delta(struct text *text, uchar *input, unsigned inlen)
{
	uchar op;
	unsigned iline, ilines, oline, olines, anchor;

	text->item = text->at;
	emit_text = input, this_text = input, end_text = input + inlen;
	emit_line = this_line = hold_lines = hold_length = 0;
	emit_op = copy_op;
patch:
	tracediff(if (listing) printf("diff %u? ", delta);)
	if (!skip_to('-') || !next_is('-') || !next_is('-') || !skip_to('\n') ||
	    !next_is('+') || !next_is('+') || !next_is('+') || !skip_to('\n') )
		goto patch;
//chunk:
	tracediff(if (listing) printf("\nchunk? ");)
	if (!next_is('@'))
		goto done;
chunk2:
	anchor = text->at - text->item;
	if (!next_is('@') || !skip_to('@') || !next_is('@') || !next_is('\n'))
		goto bad_chunk;
	/* we make the sign of the beast at whoever thought it was a good
	   idea to leave out the lines count when it is equal to one */
	{ // !!! gcc bug if no local block
	int buflen = 100;
	uchar buf[buflen], *str = text->item + anchor;
	int len = text->at - str;
	if (len > buflen)
		goto bad_chunk2;
	memcpy(buf, str, len);
	buf[len] = 0;
	if (sscanf(buf, "@ -%u,%u +%u,%u @@", &iline, &ilines, &oline, &olines) != 4) {
		trace_off(warn("variant chunk format");)
		if (sscanf(buf, "@ -%u +%u,%u @@", &iline, &oline, &olines) == 3)
			ilines = 1;
		else if (sscanf(buf, "@ -%u,%u +%u @@", &iline, &ilines, &oline) == 3)
			olines = 1;
		else if (sscanf(buf, "@ -%u +%u @@", &iline, &oline) == 2)
			{ ilines = 1, olines = 1; }
		else
			goto bad_chunk2;
	}
	}
	tracediff(if (listing) warn("chunk: %i,%i %i,%i", iline, ilines, oline, olines);)
	iline = ilines = oline = olines = 0;
	this_line = iline - 1; // zero based, no input file gives this_line = -1
	diffchunk++;
line:
	need_more(text);
	tracediff(if (listing) printf("\nline '%c' ", *text->at);)
	op = text_op;
	switch (next_ch()) {
		case '-': op++;
		case ' ': op++;
		case '+':
			anchor = text->at - text->item;
			skip_to('\n');
			if (op != emit_op)
				generate(op);
			switch (op) {
			case copy_op:
			case skip_op:
				this_line++;
				break;
			case text_op:
				if (hold_lines == maxlines) {
					trace_off(warn("expand texlines to %u", maxlines);)
					maxlines += 1000;
					expand(maxlines * sizeof(struct textline), &textlines);
				}
				uchar *start = anchor + text->item;
				unsigned length = text->at - start;
				textlines[hold_lines].source = start;
				textlines[hold_lines].length = length;
				hold_length += length;
				hold_lines++;
				break;
			}
			goto line;
		case '@': goto chunk2; // finish chunk, check num lines
		default: goto done; // finish chunk, check num lines, finish patch
	}
done:
	unnext(text);
	generate(-1);
	++delta;
	tracediff(if (listing) warn("done delta %u", delta);)
	return 0;
#if 0
again:
	generate(-1);
	return -EAGAIN;
#endif
//bad_line:
bad_chunk:
	warn("chunk header expected!");
bad_chunk2:
	hexdump(text->item + anchor - 16, 48);
	warn("at %i+%i, lim=%i top=%i size=%i", text->item - text->base, text->at - text->item, text->lim - text->base, text->top - text->base, text->chunk + text->maxline);
	warn("unrecognizable chunk %u in delta %u!  tail = %i (%i)", diffchunk, delta, text->top - (text->item + anchor), text->top - text->at);
#if 0
	char buf[100];
	pread(text->file, buf, 100, (text->filesize - text->filetail) - (text->top - text->item) + anchor - 16);
	hexdump(buf, 100);
#endif
exit(1);
	return -1;
}

/* Tridgedump parsing */

typedef unsigned atom_id_t, item_id_t, node_id_t;

struct version {
	struct list_head hash;
	struct list_head children, parents;
	struct version *parent;
	struct cstring *name;
	struct cstring *tag;
	unsigned id;
};

struct version_list { struct list_head list; struct version *version; };
struct change { struct version *parent, *child; };
struct atom { struct list_head hash; unsigned id; struct cstring name; };

static struct list_head *version_hashvec, *atom_hashvec;
static unsigned cruft, diffs, selfchanges;
static unsigned version_buckets, atom_buckets;

/* versions */

static struct memvec *version_mem;

struct version *get_version(node_id_t version)
{
	return index_item(version, version_mem);
}

#if 0
static struct dict *get_version(unsigned id)
{
	assert(is_dict(id));
	return *(struct dict **)index_item(id/2, dict_mem);
}
#endif

static void show_version_info(struct version *version)
{
	printf("%u: ", version->id);
	print_cstring(version->name);
	if (version->tag)
		printf("\"%.*s\" ", version->tag->length, version->tag->text);
}

static void show_version(struct version *version)
{
	struct list_head *list;
	list_for_each(list, &version->parents)
		print_cstring(list_entry(list, struct version_list, list)->version->name);
	printf("=> ");
	show_version_info(version);
	printf("=> ");
	list_for_each(list, &version->children)
		print_cstring(list_entry(list, struct version_list, list)->version->name);
	printf("\n");
}

static void show_versions(void)
{
	walk_memvec(version_mem, (void (*)(void *))show_version, 1);
}

static int ccount(char *str, unsigned len, char c)
{
	unsigned count = 0;
	while ((str = memchr(str, c, len)))
		str++, count++;
	return count;
}

int branchier(struct version *a, struct version *b)
{
//	return a->name->length > b->name->length;
	int dots1 = ccount(a->name->text, a->name->length, '.');
	int dots2 = ccount(b->name->text, b->name->length, '.');
//warn_("dots1 = %u, dots2 = %u", " ", dots1, dots2);
//printf("%.*s ", a->name->length, a->name->text);
//printf("%.*s\n", b->name->length, b->name->text);
	return dots1 == dots2? csless(b->name, a->name): dots1 < dots2;
#if 0
	char *p;
	unsigned segs = 0;
	while ((p = memchr(name, '.', namelen))) {
		unsigned pos = p - name + 1;
		trace_off(printf("%.*s / ", p - name, name);)
		assert(segs < maxpath - 1);
		path[segs++].atom = make_atom(name, p - name);
		name += pos;
		namelen -= pos;
	}
#endif
}

static void show_version_subtree(struct version *version, unsigned indent)
{
	unsigned i, n = 0, incols = 3;
	struct list_head *list = version->children.next;
	printf("[%.2u] %*.s", indent, indent * incols, "");
	show_version_info(version);
	printf("\n");
	list_for_each(list, &version->children)
		n++;
#if 1
	if (n) {
		struct version *vec[n];
		n = 0;
		list_for_each(list, &version->children)
			vec[n++] = list_entry(list, struct version_list, list)->version;
		COMBSORT(n, i, j, branchier(vec[i], vec[j]), { void *x = vec[i]; vec[i] = vec[j]; vec[j] = x; })
		for (i = 1; i < n; i++) {
			show_version_subtree(vec[i], indent + 1);
			if (i < n - 1) printf("[%.2u] %*.s------\n", indent, indent * incols, "");
		}
		show_version_subtree(vec[0], indent);
	}
#else
	struct list_head *first = version->children.next, *list = first->next;
	while (list != &version->children) {
		show_version_subtree(list_entry(list, struct version_list, list)->version, indent + 1);
		list = list->next;
	}
	if (first != &version->children)
		show_version_subtree(list_entry(first, struct version_list, list)->version, indent);
#endif
}

static void show_version_tree(void)
{
	show_version_subtree(get_version(0), 0);
}

static void dump_versions()
{
	unsigned i;
	for (i = 0; i < version_buckets; i++) {
		struct list_head *list;
		unsigned n = 0;
		list_for_each(list, version_hashvec + i) {
			struct version *version = list_entry(list, struct version, hash);
			show_version(version);
			n++;
			// if (!n++) printf("%u: ", i);
		}
		printf(n? "": ".");
		// if (n) printf("(%u) ", n); else printf(".");
	}
	printf("\n");
}

static struct version *find_version_in_bucket(struct cstring *name, struct list_head *bucket)
{
	struct version *version;
	struct list_head *list;
	unsigned len = name->length + 1;
	list_for_each(list, bucket)
		if (!memcmp((version = list_entry(list, struct version, hash))->name, name, len))
			return version;
	return NULL;
}

static struct version *find_version(struct cstring *name)
{
	unsigned hash = stringhash(name->text, name->length) % version_buckets;
	struct list_head *bucket = version_hashvec + hash;
	return find_version_in_bucket(name, bucket);
}

static struct version *make_version(struct cstring *name, int child)
{
	unsigned hash = stringhash(name->text, name->length) % version_buckets;
	struct list_head *bucket = version_hashvec + hash;
	struct version *version = find_version_in_bucket(name, bucket);
	if (!version) {
		unsigned id = version_mem->items;
		version = new_memitem(&version_mem);
		*version = (struct version){ .name = cstring(name->text, name->length), .id = id };
		INIT_LIST_HEAD(&version->parents);
		INIT_LIST_HEAD(&version->children);
		list_add(&version->hash, bucket);
	}
	return version;
}

void show_tags(void)
{
	int ver;
	for (ver = 0; ver < version_mem->items; ver++) {
		struct cstring *tag = get_version(ver)->tag;
		if (tag) {
			printf("%u: ", ver);
			print_cstring(tag);
			printf("\n");
		}
	}
}

/* changes */

static struct memvec *change_mem;

static void show_change(struct change *change)
{
	print_cstring(change->parent->name);
	printf("-> ");
	print_cstring(change->child->name);
	printf("\n");
}

static void show_changes()
{
	walk_memvec(change_mem, (void (*)(void *))show_change, 1);
}

static struct change *make_change(struct cstring *pname, struct cstring *cname)
{
	trace_off(warn("change %u, %.*s -> %.*s", change_mem->items, pname->length, pname->text, cname->length, cname->text);)
	struct version *parent = make_version(pname, 0);
	struct version *child = make_version(cname, 1);
	struct change *change = new_memitem(&change_mem);
	*change = (struct change){ parent, child };

	int self = !memcmp(pname, cname, cname->length + 1);
	if (self) {
		cruft(warn("change self => %.*s", *cname, cname + 1));
		selfchanges++;
	}

	if (!self) {
		struct version_list *plink = malloc(sizeof(struct version_list));
		struct version_list *clink = malloc(sizeof(struct version_list));
		clink->version = child;
		plink->version = parent;
		list_add(&plink->list, &child->parents);
		list_add(&clink->list, &parent->children);
		child->parent = parent; // !!! lose the parent list or maybe keep for merges
	}
	return change;
}

/* atoms */

static struct memvec *atom_mem;

static int show_atom(struct atom *atom)
{
	if (atom)
{
// warn("atom length = %u", atom->name.length);
// hexdump(atom->name.text, atom->name.length);
		print_cstring(&atom->name);
}
	else
		printf("<nada> ");
	return 0;
}

static void show_atomref(struct atom **atomref)
{
	show_atom(*atomref);
	printf("\n");
}

static void show_atoms()
{
	walk_memvec(atom_mem, (void (*)(void *))show_atomref, 1);
}

static void dump_atoms()
{
	unsigned i;
	struct list_head *list;
	for (i = 0; i < atom_buckets; i++)
		list_for_each(list, atom_hashvec + i)
			show_atom(list_entry(list, struct atom, hash));
	printf("\n");
}

static struct atom *find_atom(char *name, unsigned len, struct list_head *bucket)
{
	struct list_head *list;
	list_for_each(list, bucket) {
		struct atom *atom = list_entry(list, struct atom, hash);
		if (atom->name.length == len && !memcmp(&atom->name.text, name, len))
			return atom;
	}
	return NULL;
}

static struct atom *index_atom(unsigned id)
{
#if 1
	return *(struct atom **)index_item(id, atom_mem);
#else
	assert(id < atom_mem->items);
	unsigned perblock = atom_mem->perblock, block = id / perblock, offset = id - block * perblock;
	return *((struct atom **)atom_mem->block[block] + offset);
#endif
}

/* parsing */

enum linetype { CHANGESET, DIFF_FOR, DIFF_CMD, DIFF, COMMENT, TAG_COMMENT, NO_ELEMENTS };

struct tagtype { struct cstring *text; enum linetype type; };
static struct tagtypez { char *text; enum linetype type; } dumptags[] = {
	{ "ChangeSet ", CHANGESET },
	{ "diff for ", DIFF_FOR },
	{ "diff -ud ", DIFF_CMD },
	{ "--- ", DIFF },
	{ "# tag: ", TAG_COMMENT },
	{ "#", COMMENT },
	{ "no elements for seqnum ", NO_ELEMENTS },
	{ NULL }};

struct tagtype notag = { (struct cstring *)"", -1 };

enum dumpstate { GET_CHANGESET, GOT_CHANGESET, GET_DIFF_BLANK, GET_DIFF_NAMES, GET_DIFF };

struct tagtype *classify(char *line, unsigned linelen)
{
	struct tagtype *tag;
	for (tag = (struct tagtype *)dumptags; tag->text; tag++) {
		unsigned len = tag->text->length;
		if (len <= linelen && !memcmp(tag->text->text, line, len))
			return tag;
	}
	return &notag;
}

static struct atom *make_atom(char *name, unsigned len)
{
	struct list_head *bucket = atom_hashvec + stringhash(name, len) % atom_buckets;
	struct atom *atom = find_atom(name, len, bucket);
	if (!atom) {
		atom = malloc(sizeof(*atom) + len);
		memcpc(&atom->name, name, len);
		list_add(&atom->hash, bucket);
		atom->id = atom_mem->items;
		*(struct atom **)new_memitem(&atom_mem) = atom;
	}
	return atom;
}

/* versioned namespace */

struct dict {
	unsigned maxentries, entries;
	struct entry {
		struct atom *atom;
		item_id_t item;
		node_id_t birth;
		struct death {
			unsigned maxdeaths, deaths;
			node_id_t version[];
		} *death;
	} entry[]; };

static struct memvec *dict_mem;

static inline int is_dict(item_id_t item)
{
	return !(item & 1);
}

static item_id_t new_dict(void)
{
	struct dict *dict = calloc(1, sizeof(struct dict));
	item_id_t id = dict_mem->items;
	*(struct dict **)new_memitem(&dict_mem) = dict;
	trace_off(warn("new dict %p ", dict);)
	return id*2;
}

static struct dict **get_dictref(unsigned id)
{
	assert(is_dict(id));
	return (struct dict **)index_item(id/2, dict_mem);
}

static struct dict *get_dict(unsigned id)
{
	assert(is_dict(id));
	return *(struct dict **)index_item(id/2, dict_mem);
}

static unsigned count_ancestors(unsigned id)
{
	unsigned count = 0;
	struct version *version = index_item(id, version_mem);
	for (; version->parent; version = version->parent)
		count++;
	return count;
}

static node_id_t *get_ancestors(unsigned id)
{
	node_id_t n = count_ancestors(id), i;
	node_id_t *vec = malloc((n + 1) * sizeof(unsigned));
	vec[0] = n;
	struct version *version = index_item(id, version_mem);
	for (i = 1; version->parent; version = version->parent, i++)
		vec[i] = version->parent->id;
	return vec;
}

static void show_ancestors(node_id_t id, node_id_t *vec)
{
	int i, n = vec[0];
	printf("version #%u has %u ancestors, ", id, n);
	for (i = 1; i <= n; i++)
		printf("%u ", vec[i]);
	printf("\n");
}

static int is_ancestor(node_id_t id, node_id_t *vec)
{
	int i, n = vec[0];
	for (i = 1; i <= n; i++)
		if (vec[i] == id)
			return 1;
	return 0;
}

static int check_ancestor(node_id_t parent, node_id_t child)
{
	struct version *version = get_version(child);
	for (; version; version = version->parent)
		if (version->id == parent)
			return 1;
	return 0;
}

int is_live(struct entry *entry, node_id_t version)
{
	int i;
//warn("born... %i %i %i", entry->birth, version, check_ancestor(entry->birth, version));
	if (entry->birth <= version && check_ancestor(entry->birth, version)) {
//warn("born... ");
		if (entry->death)
			for (i = 0; i < entry->death->deaths; i++)
				if (check_ancestor(entry->death->version[i], version))
					return 0;
		return 1;
	}
	return 0;
} // !!! use this in a bunch of places below

static int is_empty(struct dict *dict, node_id_t version)
{
	int i;
	for (i = 0; i < dict->entries; i++)
		if (is_live(dict->entry + i, version))
			return 0;
	return 1;
}

static unsigned lookup(struct dict *dict, struct atom *atom, node_id_t version)
{
	int i;
	// printf("atom "); show_atom(atom);
	for (i = 0; i < dict->entries; i++) {
		// printf("entry "); show_atom(index_atom(entry->atom->id));
		// printf("%p = %p, %i <= %i, ancestor = %i ", entry->atom, atom, entry->birth, version, check_ancestor(entry->birth, version));
		if (dict->entry[i].atom != atom || !is_live(dict->entry + i, version))
			continue;
		// printf("found at %i\n", i); 
		return i;
	}
	// printf("not found\n"); 
	return -1;
}

static void show_dict(struct dict *dict, node_id_t version)
{
	node_id_t *ancestors = get_ancestors(version);
	trace_off(show_ancestors(id, ancestors);)

	int i, j;
	printf("in version %u ", version);
	for (i = 0; i < dict->entries; i++) {
		struct entry *entry = dict->entry + i;
		printf("entry "); show_atom(entry->atom);
		if (entry->birth <= version && check_ancestor(entry->birth, version)) {
			trace_off(printf("(%u deaths) ", entry->deaths);)
			for (j = 0; j < (entry->death? entry->death->deaths: 0); j++) {
				node_id_t death = entry->death->version[j];
				trace_off(printf("(dead at %u) ", death);)
				if (0? check_ancestor(death, version): (version == death || is_ancestor(death, ancestors))) {
					printf("is dead");
					goto silly;
				}
			}
			printf("is alive");
		} else printf("is yet to be born");
silly:		printf("; ");
	}
	printf("\n");
	free(ancestors);
}

static void list_dict(struct dict *dict, node_id_t version)
{
	int i;
//printf("entries %i ", dict->entries);
	for (i = 0; i < dict->entries; i++)
//printf("version %i, entry %i", version, is_live(dict->entry + i, version));
		if (is_live(dict->entry + i, version))
			show_atom(dict->entry[i].atom);
	printf("\n");
}

static void list_tree_(unsigned root, node_id_t version, int indent)
{
	struct dict *dict = get_dict(root);
	int i;
	for (i = 0; i < dict->entries; i++) {
		struct entry *entry = dict->entry + i;
		if (!is_live(entry, version))
			continue;
		printf("%*s", indent, "");
		show_atom(entry->atom);
		printf("%s\n", is_dict(entry->item)? "/": "");
		if (is_dict(entry->item))
			list_tree_(entry->item, version, indent + 3);
	}
}

static void list_tree(unsigned root, node_id_t version)
{
	list_tree_(root, version, 0);
}

static void show_tree_atom(struct entry *entry, unsigned indent)
{
	printf("%*s", indent, "");
	show_atom(entry->atom);
	printf("%s\n", is_dict(entry->item)? "/": "");
}

static void show_tree_(unsigned root, node_id_t version, int indent)
{
	struct dict *dict = get_dict(root);
	int i, n = 0;
	for (i = 0; i < dict->entries; i++)
		n += is_live(dict->entry + i, version);
	if (n) {
		struct entry *vec[n];
		for (n = 0, i = 0; i < dict->entries; i++)
			if (is_live(dict->entry + i, version))
				vec[n++] = dict->entry + i;
		COMBSORT(n, i, j, csless(&vec[i]->atom->name, &vec[j]->atom->name), { void *x = vec[i]; vec[i] = vec[j]; vec[j] = x; })
		for (i = 0; i < n; i++) {
			struct entry *entry = vec[i];
			show_tree_atom(entry, indent);
			if (is_dict(entry->item))
				show_tree_(entry->item, version, indent + 3);
		}
	}
}

static void show_tree(unsigned root, node_id_t version)
{
	show_tree_(root, version, 0);
}

static void dump_dict(struct dict *dict)
{
	int i;
//printf("entries %i ", dict->entries);
//	printf("\ndict ");
	for (i = 0; i < dict->entries; i++) {
		struct entry *entry = dict->entry + i;
		show_atom(entry->atom);
		printf("(%i", entry->birth);
		if (entry->death) {
			int i, n = entry->death->deaths;
			printf("-");
			for (i = 0; i < n; i++)
				printf("%i%s", entry->death->version[i], i == n-1? "": ",");
		}
		printf(") ");
	}
	printf("\n");
}

static struct entry *new_dict_entry(struct dict **dict)
{
	return new_vecitem((void **)dict, 0, sizeof(struct entry), 100);
}

static unsigned add_to_dict(node_id_t version, item_id_t dict, struct atom *atom, item_id_t item)
{
	struct dict **dictref = get_dictref(dict);
	unsigned which = (*dictref)->entries;
	*new_dict_entry(dictref) = (struct entry){ .atom = atom, .item = item, .birth = version };
	return which;
}

static node_id_t *add_death(struct death **deaths)
{
	if (!*deaths)
		*deaths = calloc(1, sizeof(struct death)); // !!! and immediately replace it
	return new_vecitem((void **)deaths, 0, sizeof(node_id_t *), 8);
}

/* Text operations */

static struct memvec *text_mem;

static inline int is_text(item_id_t item)
{
	return (item & 1);
}

static item_id_t new_text(void)
{
	item_id_t id = text_mem->items;
	*(int *)new_memitem(&text_mem) = -1;
	return id*2 + 1;
}

/* namespace operations */

#define maxpath 100

struct path { struct atom *atom; item_id_t dict; unsigned i; };

static void show_path(struct path *path, unsigned segs)
{
	//printf("path: ");
	int seg;
	for (seg = 0; seg < segs; seg++)
		show_atom(path[seg].atom);
	//printf("\n");
}

struct entry *path_entry(struct path path[], unsigned seg)
{
	return get_dict(path[seg].dict)->entry + path[seg].i;
}

static int atomize(char *name, unsigned namelen, struct path *path)
{
//return;
	trace_off(warn("%.*s", namelen, name);)
	char *p;
	unsigned segs = 0;
	while ((p = memchr(name, '/', namelen))) {
		unsigned pos = p - name + 1;
		trace_off(printf("%.*s / ", p - name, name);)
		assert(segs < maxpath - 1);
		path[segs++].atom = make_atom(name, p - name);
		name += pos;
		namelen -= pos;
	}
	trace_off(printf("%.*s (%u)\n", namelen, name, segs);)
	path[segs++].atom = make_atom(name, namelen);
	return segs;
}

static int atomize2(char *name, unsigned namelen, struct path *path, struct path *oldpath, unsigned *same)
{
	int segs = atomize(name, namelen, path), seg, check = *same;
	// can be optimized by using string compares on segs instead of atom hash
	if (check > segs)
		check = segs;
	for (*same = 0, seg = 0; seg < check; (*same)++, seg++)
		if (path[seg].atom != oldpath[seg].atom)
			break;
	return segs;
}

static int walk_path(struct path *path, unsigned segs, item_id_t dictid, node_id_t version, int make)
{
	int seg = 0;
	while (1) {
		int final = (seg == segs - 1);
		struct atom *atom = path[seg].atom;
		path[seg].dict = dictid;
		if ((path[seg].i = lookup(get_dict(dictid), atom, version)) == -1) {
			if (!make)
				return seg;
			names(warn_("create %s", " ", final? "text": "dict"); show_path(path, seg + 1); printf("in version %u\n", version);)
			item_id_t item = final? new_text(): new_dict();
			path[seg].i = add_to_dict(version, dictid, atom, item);
		}
		item_id_t id = get_dict(dictid)->entry[path[seg].i].item;
		if (final/*&& is_text(id)*/)
			return segs;
		if (is_text(id))
{
error("expecting dict");
			return seg;
}
		dictid = id; // but we already have this!
		seg++;
	}
}

/* Current version */

static unsigned current;

static void show_current_dict(struct dict **dict)
{
//	warn("dict %p entries = %i", *dict, (*dict)->entries);
//	show_dict(*dict, current);
//	list_dict(*dict, current);
	dump_dict(*dict);
}

static void show_dicts(void)
{
	walk_memvec(dict_mem, (void (*)(void *))show_current_dict, 1);
}

static int change_header(char *cmd, int lim, struct change **changeref)
{
	char cpname[lim + 1], ccname[lim + 1], eol;
	char *pname = cpname + 1, *cname = ccname + 1;
	int vals = sscanf(cmd, "ChangeSet (%[^)]) -> ChangeSet (%[^)])%c", pname, cname, &eol);
	if (vals != 2)
		return -EINVAL;
	*cpname = strlen(cpname + 1);
	*ccname = strlen(ccname + 1);
	struct change *change = *changeref = make_change((struct cstring *)cpname, (struct cstring *)ccname);
	trace_off(warn_("parent %u:%.*s", " ", change->parent->id, *cpname, cpname + 1);)
	trace_off(printf("child %u:%.*s\n", change->child->id, *cpname, cpname + 1);)
	trace_off(warn("[%u] %s", change_mem->items, cmd);)
	return change->child->id;
}

static int load_spdump(struct text *text, item_id_t root)
{
	uchar *next;
	unsigned maxcommand = 255 + 1;
	char cmd[maxcommand];
	enum dumpstate state = GET_CHANGESET;
	get_more(text);
	struct change *change = NULL;
	while (text->top - text->at + text->filetail) {
		ltrace(show_line();)
		if (!(next = memchr(text->at, '\n', text->top - text->at))) {
			trace(warn("find needs more, tail = %Lu", text->filetail);)
			text->item = text->at;
			get_more(text);
			continue;
		}
		unsigned linelen = next - text->item;
		struct tagtype *hit = classify(text->at, linelen);
		enum linetype linetype = hit->type;
		unsigned pos = hit->text->length;
		//warn(">>> classify line [%.*s] as %u", linelen, text->at, linetype);
		//listing = inlines >= 3000 && 3500 >= inlines;
jump:
		switch (state) {
		case GET_CHANGESET:
			if (linetype == CHANGESET) {
				cruft(warn("%.*s", linelen, text->at);)
				if (linelen < maxcommand) {
					unsigned id;
					memcpz(cmd, text->at, linelen);
					if ((id = change_header(cmd, maxcommand, &change)) < 0)
						warn("Changeset syntax error at line %u", line());
					current = id;
#if 0
if (current == 1156) {
	list_tree(root, current);
	exit(1);
}
#endif
					trace_off(warn("current = %u", current);)
				}
				break;
			}
			if (linetype == TAG_COMMENT) {
				tags(warn("tag version %u as [%.*s]", current, linelen - pos, text->at + pos);)
				if (change) {
					if (change->child->tag)
						warn("version %.*s already tagged!", change->child->name->length, change->child->name->text);
					change->child->tag = cstring(text->at + pos, linelen - pos);
				}
				break;
			}
			if (linetype == COMMENT) {
				tracedump(warn("comment [%.*s]", linelen, text->at);)
				break;
			}
			if (linetype == NO_ELEMENTS) {
				unsigned num;
				getu(text->at + pos, linelen - pos, &num);
				cruft(warn("no elements for seqnum %u", num);)
				cruft++;
				break;
			}
			if (linetype == DIFF_FOR) {
				state = GET_DIFF;
				goto jump;
			}
			break;
		case GET_DIFF_BLANK:
			tracedump(warn("empty line [%.*s]", linelen, text->at);)
			state = GET_DIFF_NAMES;
			break;
		case GET_DIFF_NAMES:
			if (linetype == DIFF_CMD) {
				text->at += pos;
				linelen -= pos;
				uchar *sep;
				if (!(sep = memchr(text->at, ' ', linelen)))
					break;
				uchar *old = text->at, *new = sep + 1;
				unsigned oldlen = sep - text->at, newlen = linelen - oldlen - 1;
				//warn("old [%.*s]", oldlen, old);
				//warn("new [%.*s]", newlen, new);
				struct path oldpath[maxpath];
				struct path newpath[maxpath];
				int oldsegs = atomize(old, oldlen, oldpath), same = oldsegs;
				int newsegs = atomize2(new, newlen, newpath, oldpath, &same);
				trace_off(printf("old"); show_path(oldpath, oldsegs);)
				trace_off(printf("new"); show_path(newpath, newsegs);)
				trace_off(warn("same = %u", same);)
				/*
				 * if oldsegs, newsegs and same not equal, then it is a move:
				 *    if the new path starts with Bitkeeper/deleted then just kill the new path
				 *       item, but remember the mangled delete name in case an undelete shows up
				 *    if the old path starts with Bitkeeper/deleted then it is an undelete: look
				 *       for the mangled delete name to find the item to create on the new path
				 *    walk oldpath to find entry to kill, and possibly remove empty dicts
				 *    walk newpath to create new entry, and possibly create new dicts
				 * Even if it is not a move, it may create a new item and possibly dicts
				 * Deletion can be left as a move for now, but empty dicts still need to die.
				 */
				trace_off(warn("names [%.*s]", linelen, text->at);)
				if (0 && (newlen != oldlen || memcmp(new, old, oldlen))) {
					char *bkdel = "BitKeeper/deleted/", bklen = strlen(bkdel);
					if (oldlen > bklen && !memcmp(old, bkdel, bklen))
						warn("?!? undelete %.*s ?!?", newlen, new);
					else if (newlen > bklen && !memcmp(new, bkdel, bklen))
						trace_off(warn("delete %.*s", oldlen, old));
				}
				if (oldsegs != newsegs || oldsegs != same) {
					names(warn("at line %u in version %u move [%.*s]", line(), current, linelen, text->at);)
					int found;
					if ((found = walk_path(oldpath, oldsegs, root, current, 0)) != oldsegs) {
						cruft(warn("at line %u entry %u missing: [%.*s]", line(), found, linelen, text->at);)
						cruft++;
						state = GET_CHANGESET;
						break;
					}
					int seg = oldsegs - 1;
					do {
						names(warn_("kill %s", " ", seg < oldsegs - 1? "dict": "text"); show_path(oldpath, seg + 1); printf("in version %u\n", current);)
						*add_death(&path_entry(oldpath, seg)->death) = current;
					} while (is_empty(get_dict(oldpath[seg].dict), current) && seg--);
				}

				if (walk_path(newpath, newsegs, root, current, 1) != newsegs) {
					warn("entry create failed");
					break;
				}
				// remember the new item for text and attribute changes
				state = GET_DIFF;
			}
			break;
		case GET_DIFF:
			if (linetype == CHANGESET) {
				state = GET_CHANGESET;
				goto jump;
			}
			if (linetype == COMMENT) {
				tracedump(warn("comment [%.*s]", linelen, text->at);)
				break;
			}
			if (!linelen) {
				tracedump(warn("<blank>");)
				break;
			}
			if (linetype == DIFF_FOR) {
				tracedump(warn("[%.*s]", linelen, text->at);)
				state = GET_DIFF_BLANK;
				break;
			}
			if (linetype == DIFF) {
//				text->at = next;
				unsigned outsize = 100;
				outbuf = output = malloc(outsize);
				outlim = outbuf + outsize;
				trace_off(warn("read a diff");)
				diff2delta(text, "", 0);
				diffs++;
				trace_off(warn("delta length %u", (char *)output - (char *)outbuf);)
				free(outbuf);
				text->item = text->at;
				need_more(text);
				state = GET_CHANGESET;
				continue;
			}
			cruft(warn("expecting comment, diff or changeset at line %u, not \"%.*s\"", line(), linelen, text->at);)
			cruft++;
			break;
		default:
			warn("unknown line type [%.*s]", linelen, text->at);
		}
		inlines++;
		text->at = next + 1;
		text->item = text->at;
		need_more(text);
	}
	tracediff(printf("\n");)
	return 0;
#if 0
read_error:
	error("Read error %i, %s", errno, strerror(errno));
	return -1;
#endif
}

static int atomiz(char *str, struct path *path)
{
	return atomize(str, strlen(str), path);
}

int main(int argc, char *argv[])
{
#if 0
	printf("less %i\n", less(cstrz("bb"), cstrz("b")));
	exit(1);
#endif
#if 0
	void *foo = NULL;
	if (posix_memalign(&foo, 1 << 9, 1 << 20))
		error("memalign error %s (%u)", strerror(errno), errno);
	warn("foo = %p", foo);
	return 0;
#endif
//	if (argc < 2)
//		error("Usage: %s filename", argv[0]);

#if 0
	version_mem = new_memvec(1000, sizeof(struct version), "versions");
	change_mem = new_memvec(1000, sizeof(struct change), "changes");
	atom_mem = new_memvec(1000, sizeof(struct atom *), "atoms");
	version_hashvec = new_hashvec(version_buckets = 0xb);
	atom_hashvec = new_hashvec(atom_buckets = 3);
	void *free1, *free2;

	make_change(free1 = cstrz("1.0"), free2 = cstrz("1.1")); free(free1); free(free2);
	make_change(free1 = cstrz("1.1"), free2 = cstrz("1.2")); free(free1); free(free2);
	make_change(free1 = cstrz("1.1"), free2 = cstrz("1.1.1")); free(free1); free(free2);
	make_change(free1 = cstrz("1.1.1"), free2 = cstrz("1.1.2")); free(free1); free(free2);
	show_versions();

	trace_off(warn("ancestors = %u", count_ancestors(find_version(cstrz("1.2"))->id));)

	struct atom *atom0 = make_atom("foo", 3);
	struct atom *atom1 = make_atom("bar", 3);

	dict_mem = new_memvec(200, sizeof(struct dict *), "dicts");
	text_mem = new_memvec(200, sizeof(int), "texts");
	item_id_t root = new_dict();

	struct path path[maxpath];
	int segs, walked;

	segs = atomiz("a/ab/abc1", path); show_path(path, segs);
	walked = walk_path(path, segs, root, 0, 1);

	segs = atomiz("a/bb/abc2", path); show_path(path, segs);
	walked = walk_path(path, segs, root, 0, 1);

	*add_death(&path_entry(path, segs - 1)->death) = 1;

	segs = atomiz("a/ab/abc3", path); show_path(path, segs);
	walked = walk_path(path, segs, root, 0, 1);
	show_dicts();
	segs = atomiz("a2", path); show_path(path, segs);
	walked = walk_path(path, segs, root, 0, 1);
	show_dicts();

	segs = atomiz("a/ab/abc3", path); show_path(path, segs);
	walked = walk_path(path, segs, root, 0, 0);

	*add_death(&path_entry(path, segs - 1)->death) = 1;

	show_dicts();
	list_tree(root, 0);
	list_tree(root, 1);
	warn("empty? %i", is_empty(get_dict(3*2 + 1), 0));
	warn("empty? %i", is_empty(get_dict(3*2 + 1), 1));
	exit(0);

	for (current = 0; current < 4; current++) {
		warn("version %u", current);
		list_tree(root, 0);
	}
	exit(0);
#endif
	char *name = argc == 1? "ccache.sp": argv[1];
	struct text text = { };

	if ((text.file = open(name, O_RDONLY | (0? O_DIRECT: 0))) == -1)
		error("%s? %s", name, strerror(errno));
	if ((text.direct = open(name, O_RDONLY | O_DIRECT)) == -1)
		error("%s? %s", name, strerror(errno));
	struct stat stat;
	if (fstat(text.file, &stat) == -1)
		error("stat error %s", strerror(errno));

	text.filetail = text.filesize = stat.st_size;
	text.chunk = 1 << 20;
	text.maxline = 1 << 8;
	text.align = 12;
	text.base = text.item = text.at = text.lim = text.top = 0;
//malloc(text.chunk + text.maxline);

	version_mem = new_memvec(1000, sizeof(struct version), "versions");
	change_mem = new_memvec(1000, sizeof(struct change), "changes");
	atom_mem = new_memvec(1000, sizeof(struct atom *), "atoms");
	version_hashvec = new_hashvec(version_buckets = ((text.filesize >> 20) + 17) | 0xb);
	atom_hashvec = new_hashvec(atom_buckets = 1023);
	dict_mem = new_memvec(200, sizeof(struct dict *), "dicts");
	text_mem = new_memvec(200, sizeof(int), "texts");

	maxlines = 30;
	textlines = malloc(maxlines * sizeof(struct textline));

	struct tagtype *tag;
	for (tag = (struct tagtype *)dumptags; tag->text; tag++)
		tag->text = cstrz((char *)tag->text);

	verbose(warn("undump %s (%Lu bytes)", name, (long long)text.filetail);)
	verbose(warn("version hash buckets = %u", version_buckets);)
	item_id_t root = new_dict();
	int err = load_spdump(&text, root);
//	show_memvec("change", change_mem);
//	show_memvec("version", version_mem);
//	show_changes();
//	dump_versions();
//	show_versions();
//	show_atoms();
//	dump_atoms();
	int arg;
	stats(warn(
		"%s: %u lines, %u versions, %u changes, %u diffs, %u atoms, %u cruftons",
		name, inlines, version_mem->items, change_mem->items - selfchanges,
		diffs, atom_mem->items, cruft);)

	for (arg = 2; arg < argc; arg++) {
		int ver = atoi(argv[arg]), versions = version_mem->items;
		if (ver < 0)
			ver = versions + ver;
		if (ver < 0)
			ver = 0;
		printf("--- %i ---\n", ver);
		if (ver >= version_mem->items) {
			warn("No such version (0-%u)", version_mem->items - 1);
			continue;
		}
		show_tree(root, ver);
	}
//show_tags();
//show_version_tree();
	exit(err); // otherwise valgrind thinks the hash table got lost (??)
	return err;
}

void *undump_useme[] = {
	atomiz, lcount_text, show_line, show_memvec, find_version,
	dump_versions, show_versions, show_changes, show_atoms, dump_atoms,
	list_dict, show_dict, show_dicts, list_tree, show_version_tree,
	show_ancestors, show_path, alter, transcheck, index_atom, };
